/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "SourceLine.h"
#include "mem_limits.h"
#include "token.h"
#include "tokens/Tokenizer.h"
#include "tokens/StringKeyHash.h"
#include "glearray.h"
#include "polish.h"
#include "pass.h"
#include "var.h"
#include "gprint.h"
#include "cutils.h"

#define true (!false)
#define false 0

/* hash tables for the global and local variables */

GLEVarMap  g_VarGlobal;
GLEVarMap* g_VarLocal = NULL;

/* global vars */

vector<double> var_val;
vector<string> var_str;

/* local strings and values*/

GLELocalVars* local_var = NULL;
vector<GLELocalVars*> local_var_stack;
int local_var_stack_level = 0;

GLELocalVars::GLELocalVars(int num) : var_val(num, 0.0), var_str(num, "") {
}

GLELocalVars::~GLELocalVars() {
}

void GLELocalVars::expand(int num) {
	if (var_val.size() <= num) {
		var_val.resize(num+1, 0.0);
		var_str.resize(num+1, "");
	}
}

void GLELocalVars::copyFrom(GLELocalVars* other, int nb) {
	expand(nb);
	for (int i = 0; i < nb; i++) {
		var_val[i] = other->var_val[i];
		var_str[i] = other->var_str[i];
	}
}

GLELocalVars* GLELocalVars::clone(int nb) {
	GLELocalVars* res = new GLELocalVars(nb);
	res->copyFrom(this, nb);
	return res;
}

void GLELocalVars::copyFrom(GLELocalVars* other) {
	copyFrom(other, other->size());
}

GLEVarSubMap::GLEVarSubMap(GLEVarMap* parent) {
	m_Parent = parent;
}

GLEVarSubMap::~GLEVarSubMap() {
}

void GLEVarSubMap::clear() {
	m_Map.clear();
	m_Idxs.clear();
}

void GLEVarSubMap::var_add(const string& name, int idx) {
	m_Map.add_item(name, idx);
	m_Idxs.push_back(idx);
}

void GLEVarSubMap::removeFromParent() {
	for (int i = 0; i < m_Idxs.size(); i++) {
		m_Parent->removeVar(m_Idxs[i]);
	}
}

GLEVarMap::GLEVarMap() {
	m_IsTemp = false;
}

GLEVarMap::~GLEVarMap() {
	clear();
}

int GLEVarMap::getFreeID() {
	if (m_Free.size() > 0) {
		int id = m_Free.back();
		m_Free.pop_back();
		return id;
	} else {
		return -1;
	}
}

int GLEVarMap::addVarIdx(const string& name) {
	int idx = getFreeID();
	int type = str_var(name) ? 2 : 1;
	if (idx != -1) {
		m_Names[idx] = name;
		m_Types[idx] = type;
	} else {
		idx = m_Names.size();
		m_Names.push_back(name);
		m_Types.push_back(type);
	}
	return idx;
}

int GLEVarMap::var_find_add(const string& name, bool* isnew) {
	*isnew = false;
	int idx = m_Map.try_get(name);
	if (idx == -1) {
		idx = addVarIdx(name);
		m_Map.add_item(name, idx);
		*isnew = true;
	}
	return idx;
}

int GLEVarMap::var_find_add_submap(const string& name, bool* isnew) {
	*isnew = false;
	GLEVarSubMap* sub = m_SubMap.back();
	int idx = sub->var_get(name);
	if (idx == -1) {
		idx = addVarIdx(name);
		sub->var_add(name, idx);
		*isnew = true;
	}
	return idx;
}

int GLEVarMap::var_get(const string& name) {
	int idx = -1;
	int submap = m_SubMap.size()-1;
	while (submap >= 0 && (idx = m_SubMap[submap]->var_get(name)) == -1) {
		submap--;
	}
	if (idx != -1) return idx;
	else return m_Map.try_get(name);
}

const string& GLEVarMap::var_name(int idx) {
	return m_Names[idx];
}

void GLEVarMap::list() {
	for (int i = 0; i < m_Names.size(); i++) {
		if (m_Types[i] != -1) {
			cout << m_Names[i] << " (" << i << ")" << endl;
		}
	}
}

void GLEVarMap::removeVar(int idx) {
	m_Free.push_back(idx);
	m_Names[idx] = "?";
	m_Types[idx] = -1;
}

GLEVarSubMap* GLEVarMap::pushSubMap() {
	GLEVarSubMap* sub = new GLEVarSubMap(this);
	m_SubMap.push_back(sub);
	return sub;
}

void GLEVarMap::popSubMap() {
	GLEVarSubMap* sub = m_SubMap.back();
	sub->removeFromParent();
	delete sub;
	m_SubMap.pop_back();
}

void GLEVarMap::clearSubMaps() {
	for (int i = 0; i < m_SubMap.size(); i++) {
		delete m_SubMap[i];
		m_SubMap[i] = NULL;
	}
	m_SubMap.clear();
}

void GLEVarMap::clear() {
	m_Names.clear();
	m_Map.clear();
	clearSubMaps();
}

void expand_global_vars(int max) {
	string init_str = "";
	while (var_val.size() <= max) {
		var_val.push_back(0.0);
		var_str.push_back(init_str);
	}
}

bool var_check(int *j) {
	int var = *j;
	/* convert var index and return true if var is local */
	if (GLE_VAR_IS_LOCAL(var)) {
		var &= ~GLE_VAR_LOCAL_BIT;
		if (g_VarLocal == NULL) {
			gprint("No local variables assigned");
			*j = 0;
		} else if (var < 0 || var >= g_VarLocal->size() || var >= NUM_LOCAL) {
			gprint("Local variable index out of range: %d is not in 0-%d", var, g_VarLocal->size());
			*j = 0;
		} else {
			*j = var;
			return true;
		}
	} else {
		if (var < 0 || var >= g_VarGlobal.size()) {
			gprint("Global variable index out of range: %d is not in 0-%d", var, g_VarGlobal.size());
			*j = 0;
		}
	}
	return false;
}

GLELocalVars* get_local_vars() {
	return local_var;
}

void var_alloc_local(int num) {
	local_var_stack_level++;
	if (local_var_stack_level >= local_var_stack.size()) {
		if (local_var_stack_level == 1) {
			// Initialize first level
			local_var_stack.push_back(NULL);
		}
		local_var = new GLELocalVars(num);
		local_var_stack.push_back(local_var);
	} else {
		local_var = local_var_stack[local_var_stack_level];
		local_var->expand(num);
	}
}

void var_alloc_local(GLEVarMap* map) {
	var_alloc_local(map->size());
}

void var_free_local() {
	if (local_var_stack_level == 0) {
		cerr << "GLE internal error: too many pops of local variable stack" << endl;
		exit(1);
	}
	local_var_stack_level--;
	local_var = local_var_stack[local_var_stack_level];
}

void var_set(int jj, double v) {
	if (var_check(&jj)) {
		local_var->var_val[jj] = v;
	} else {
		var_val[jj] = v;
	}
}

void var_setstr(int jj, char *s) {
	if (var_check(&jj)) {
		local_var->var_str[jj] = s;
	} else {
		var_str[jj] = s;
	}
}

void var_getstr(int jj, char *s) {
	const char* str = NULL;
	if (var_check(&jj)) {
		str = local_var->var_str[jj].c_str();
	} else {
		str = var_str[jj].c_str();
	}
	strcpy(s, str);
}

void var_getstr(int jj, string& s) {
	if (var_check(&jj)) {
		s = local_var->var_str[jj];
	} else {
		s = var_str[jj];
	}
}

void var_get(int jj, double *v) {
	if (var_check(&jj)) {
		*v = local_var->var_val[jj];
	} else {
		*v = var_val[jj];
	}
}

char* var_get_name(int jj) {
	if (var_check(&jj)) {
		return (char*)g_VarLocal->var_name(jj).c_str();
	} else {
		return (char*)g_VarGlobal.var_name(jj).c_str();
	}
}

void var_nlocal(int *l) {
	*l = g_VarLocal == NULL ? 0 : g_VarLocal->size();
}

void var_clear_global() {
	g_VarGlobal.clear();
}

GLEVarMap* var_swap_local_map(GLEVarMap* map) {
	GLEVarMap* old = g_VarLocal;
	g_VarLocal = map;
	return old;
}

void var_set_local_map(GLEVarMap* map) {
	g_VarLocal = map;
}

void var_clear_local() {
	g_VarLocal = NULL;
}

bool has_local_var_map() {
	return g_VarLocal != NULL;
}

GLEVarMap* get_local_var_map() {
	return g_VarLocal;
}

void var_init(int idx, int type) {
	if (type == 1) var_set(idx, 0.0);
	else var_setstr(idx, "");
}

void var_def(char *s, double x) {
	int idx, type=1;
	var_findadd(s, &idx, &type);
	var_set(idx, x);
}

void var_findadd_set(char* name, double value) {
	int idx, type = 1;
	var_findadd(name, &idx, &type);
	var_set(idx, value);
}

void var_add_local(char *name, int *idx, int *type) {
	bool isnew;
	int l_idx = g_VarLocal->var_find_add(name, &isnew);
	*type = g_VarLocal->getType(l_idx);
	*idx = l_idx | GLE_VAR_LOCAL_BIT;
}

void var_findadd(char *name, int *idx, int *type) {
	bool isnew;
	if (g_VarLocal != NULL && g_VarLocal->hasSubMap()) {
		// cout << "adding to submap: " << name << endl;
		int l_idx = g_VarLocal->var_find_add_submap(name, &isnew);
		*type = g_VarLocal->getType(l_idx);
		*idx = l_idx | GLE_VAR_LOCAL_BIT;
		local_var->expand(l_idx);
		if (isnew) {
			var_init(*idx, *type);
		}
	} else {
		if (g_VarLocal != NULL) {
			// cout << "adding to local: " << name << endl;
			int l_idx = g_VarLocal->var_get(name);
			if (l_idx != -1) {
				*type = g_VarLocal->getType(l_idx);
				*idx = l_idx | GLE_VAR_LOCAL_BIT;
				return;
			}
		}
		*idx = g_VarGlobal.var_find_add(name, &isnew);
		*type = g_VarGlobal.getType(*idx);
		// cout << "adding to global: " << name << " idx = " << *idx << " type = " << *type << " new = " << isnew << endl;
		if (isnew) {
			expand_global_vars(*idx);
			var_init(*idx, *type);
		}
	}
}

// First search in local vars, then in global vars
void var_find(char *name, int *idx, int *type) {
	*idx = -1;
	if (g_VarLocal != NULL) {
		int l_idx = g_VarLocal->var_get(name);
		if (l_idx != -1) {
			*type = g_VarLocal->getType(l_idx);
			*idx = l_idx | GLE_VAR_LOCAL_BIT;
			return;
		}
	}
	int g_idx = g_VarGlobal.var_get(name);
	if (g_idx != -1) {
		*type = g_VarGlobal.getType(g_idx);
		*idx = g_idx;
	}
}

GLEVarSubMap* var_add_local_submap() {
	if (g_VarLocal == NULL) {
		g_VarLocal = new GLEVarMap();
		g_VarLocal->setTemp(true);
		var_alloc_local(0);
	}
	return g_VarLocal->pushSubMap();
}

void var_remove_local_submap() {
	if (g_VarLocal != NULL) {
		if (g_VarLocal->isTemp()) {
			delete g_VarLocal;
			g_VarLocal = NULL;
			var_free_local();
		} else {
			g_VarLocal->popSubMap();
		}
	}
}

void var_find_dn(GLEVarSubMap* map, int *idx, int *var, int *nd) {
	*nd = 0;
	for (int i = 0; i < map->size(); i++) {
		int lidx = map->get(i);
		const string& name = g_VarLocal->var_name(lidx);
		if (str_ni_equals(name.c_str(), "D", 1)) {
			int d = atoi(name.c_str()+1);
			// cout << "name = " << name << " idx = " << d << endl;
			if (d > 0 && d < 100) {
				*idx = lidx | GLE_VAR_LOCAL_BIT;
				*var = d;
				(*nd)++; idx++; var++;
			}
		}
	}
	// cout << "res = " << *nd << endl;
}

bool str_var(const char* s) {
	int i = strlen(s);
	if (*(s+i-1)=='$') return true;
	else return false;
}

bool str_var(const string& s) {
	if (s[s.length()-1] == '$') return true;
	else return false;
}

int valid_var(char* s) {
	return true;
}

bool var_valid_name(const string& name) {
	if (name.length() == 0) {
		return false;
	} else {
		int ch = name[0];
		if (!((ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z'))) {
			return false;
		} else {
			return true;
		}
	}
}
